import os
import io
from typing import Optional
import time

import logging

from pytubefix.exceptions import AgeRestrictedError

from warp_beacon.jobs.types import JobType
from warp_beacon.scraper.youtube.abstract import YoutubeAbstract

from warp_beacon.scraper.exceptions import NotFound, YotubeAgeRestrictedError

class YoutubeShortsScraper(YoutubeAbstract):
	YT_MAX_RETRIES_DEFAULT = 8
	YT_PAUSE_BEFORE_RETRY_DEFAULT = 3
	YT_TIMEOUT_DEFAULT = 2
	YT_TIMEOUT_INCREMENT_DEFAULT = 60

	def _download(self, url: str, session: bool = True, thumbnail: Optional[io.BytesIO] = None, timeout: int = 60) -> list:
		res = self._download_pytubefix_max_res(url=url, session=session, thumbnail=thumbnail, timeout=timeout)
		if not res:
			res = self._download_pytube_dash(url=url, session=session, thumbnail=thumbnail, timeout=timeout)

		return res

	def _download_pytube_dash(self, url: str, session: bool = True, thumbnail: Optional[io.BytesIO] = None, timeout: int = 0) -> list:
		res = []
		yt = self.build_yt(url, session=session)
		stream = yt.streams.get_highest_resolution()

		if not stream:
			raise NotFound("No suitable video stream found")

		local_file = stream.download(
			output_path=self.DOWNLOAD_DIR,
			max_retries=0,
			timeout=timeout,
			skip_existing=False,
			filename_prefix="yt_download_"
		)

		local_file = self.rename_local_file(local_file)

		logging.debug("Temp filename: '%s'", local_file)
		res.append({
			"local_media_path": local_file,
			"performer": yt.author,
			"thumb": thumbnail,
			"canonical_name": stream.title,
			"media_type": JobType.VIDEO
		})

		return res

	def _download_yt_dlp(self, url: str, thumbnail: Optional[io.BytesIO] = None, timeout: int = 60) -> list:
		res = []
		with self.build_yt_dlp(timeout) as ydl:
			info = ydl.extract_info(url, download=True)
			local_file = ydl.prepare_filename(info)
			logging.debug("Temp filename: '%s'", local_file)
			res.append({
				"local_media_path": local_file,
				"performer": info.get("uploader", "Unknown"),
				"thumb": thumbnail,
				"canonical_name": info.get("title", ''),
				"media_type": JobType.VIDEO
			})

		return res
	
	def _download_pytubefix_max_res(self, url: str, session: bool = True, thumbnail: Optional[io.BytesIO] = None, timeout: int = 60) -> list:
		res = []
		local_video_file, local_audio_file = '', ''
		try:
			yt = self.build_yt(url, session=session)
			
			video_stream = yt.streams.filter(adaptive=True, file_extension='mp4', only_video=True).order_by('resolution').desc().first()
			audio_stream = yt.streams.filter(adaptive=True, file_extension='mp4', only_audio=True).order_by('abr').desc().first()

			local_video_file = video_stream.download(
				output_path=self.DOWNLOAD_DIR,
				max_retries=3,
				timeout=timeout,
				skip_existing=False,
				filename_prefix="yt_download_video_"
			)
			local_video_file = self.rename_local_file(local_video_file)
			logging.debug("Temp video filename: '%s'", local_video_file)
			local_audio_file = audio_stream.download(
				output_path=self.DOWNLOAD_DIR,
				max_retries=3,
				timeout=timeout,
				skip_existing=False,
				filename_prefix="yt_download_audio_"
			)
			local_audio_file = self.rename_local_file(local_audio_file)
			logging.debug("Temp audio filename: '%s'", local_audio_file)

			muxed_video = self.mux_raw_copy(
				video_path=local_video_file,
				audio_path=local_audio_file,
				output_path=f"{self.DOWNLOAD_DIR}/yt_muxed_video_{int(time.time())}.mp4")
			if muxed_video:
				muxed_video = self.rename_local_file(muxed_video)
				logging.debug("Temp muxed filename: '%s'", muxed_video)

				res.append({
					"local_media_path": muxed_video,
					"performer": yt.author,
					"thumb": thumbnail,
					"canonical_name": yt.title,
					"media_type": JobType.VIDEO
				})
		except AgeRestrictedError as e:
			raise YotubeAgeRestrictedError("Youtube Age Restricted error")
		finally:
			if os.path.exists(local_video_file):
				os.unlink(local_video_file)
			if os.path.exists(local_audio_file):
				os.unlink(local_audio_file)
		
		return res


import logging
from typing import Callable, List, Union, Optional
from copy import deepcopy
from urllib.parse import urlparse, parse_qs, urlencode, urlunparse
from pathlib import Path
import time
from pydantic_core import ValidationError
import requests

from instagrapi import Client
from instagrapi.types import Media, User, Story
from instagrapi.exceptions import (
	ClientError,
	#ClientLoginRequired,
	ClientNotFoundError,
	VideoNotDownload,
	MediaNotFound,
	PrivateError
)

from instagrapi.extractors import (
	#extract_location,
	extract_media_v1,
	#extract_user_short,
	#extract_usertag,
	#extract_resource_v1
)


def safe_extract_media_v1(data: dict) -> Media:
	try:
		return extract_media_v1(data)
	except ValidationError as e:
		if "clips_metadata.original_sound_info" in str(e):
			logging.info("clips_metadata.original_sound_info handling")
			cm = data.get("clips_metadata")
			if isinstance(cm, dict) and cm.get("original_sound_info") is not None:
				cm.pop("original_sound_info", None)
			elif isinstance(cm, dict) and cm.get("original_sound_info") is None:
				cm.pop("original_sound_info", None)
			return extract_media_v1(data)
		raise

'''
def extract_media_v1(data):
	"""Extract media from Private API"""
	media = deepcopy(data)
	if "video_versions" in media:
		# Select Best Quality by Resolutiuon
		media["video_url"] = sorted(
			media["video_versions"], key=lambda o: o["height"] * o["width"]
		)[-1]["url"]
	if media["media_type"] == 2 and not media.get("product_type"):
		media["product_type"] = "feed"
	if "image_versions2" in media:
		media["thumbnail_url"] = sorted(
			media["image_versions2"]["candidates"],
			key=lambda o: o["height"] * o["width"],
		)[-1]["url"]
	if media["media_type"] == 8:
		# remove thumbnail_url and video_url for albums
		# see resources
		media.pop("thumbnail_url", "")
		media.pop("video_url", "")
	location = media.get("location")
	media["location"] = location and extract_location(location)
	media["user"] = extract_user_short(media.get("user"))
	media["usertags"] = sorted(
		[
			extract_usertag(usertag)
			for usertag in media.get("usertags", {}).get("in", [])
		],
		key=lambda tag: tag.user.pk,
	)
	media["like_count"] = media.get("like_count", 0)
	media["has_liked"] = media.get("has_liked", False)
	#media["sponsor_tags"] = [tag["sponsor"] for tag in media.get("sponsor_tags", [])]
	media["sponsor_tags"] = [tag["sponsor"] for tag in media.get("sponsor_tags") or []]
	media["play_count"] = media.get("play_count", 0)
	media["coauthor_producers"] = media.get("coauthor_producers", [])
	return Media(
		caption_text=(media.get("caption") or {}).get("text", ""),
		resources=[
			extract_resource_v1(edge) for edge in media.get("carousel_media", [])
		],
		**media,
	)
'''
class WBClient(Client):
	"""
	patched instagrapi
	"""
	def __init__(self, user_agent: str) -> None:
		super().__init__()
		self.progress_callback = None
		self.__session = requests.Session()
		#self.__session.headers["User-Agent"] = user_agent
		self.__session.headers.update({
			#"User-Agent": ScraperUtils.get_ua(),
			"User-Agent": user_agent,
			"Accept": "*/*",
			"Accept-Language": f"{self.settings['locale']},{self.settings['country']};q=0.8",
			"Accept-Encoding": "gzip, deflate",
			#"Referer": "https://www.instagram.com/",
			"Connection": "keep-alive",
			#"Sec-Fetch-Site": "same-origin",
			#"Sec-Fetch-Mode": "navigate",
			#"Sec-Fetch-User": "?1",
			#"Sec-Fetch-Dest": "document",
			#"Upgrade-Insecure-Requests": "1",
			#"DNT": "1",
		})
		#logging.info("Builded UA: '%s'", self.__session.headers)
		self.essential_params = {"oe", "oh", "_nc_ht", "_nc_cat", "_nc_oc", "_nc_ohc", "_nc_gid"}

	def set_progress_callback(self, callback: Callable[[int | None, int, Path], None]) -> None:
		if not callback or not callable(callback):
			raise TypeError("Progress callback must be callable")
		self.progress_callback = callback

	def adaptive_chunk_size(self, content_length: int) -> int:
		if content_length < 100_000:
			return 2048
		elif content_length < 5_000_000:
			return 8192
		elif content_length < 100_000_000:
			return 32768
		else:
			return 65536

	def sanitize_instagram_url(self, url: str) -> str:
		if "oh=" in url: # signed url, do not touch
				return url
		parsed = urlparse(url)
		query = parse_qs(parsed.query)
		filtered_query = {k: v for k, v in query.items() if k in self.essential_params}
		new_query = urlencode(filtered_query, doseq=True)
		return urlunparse(parsed._replace(query=new_query))

	def video_download_by_url(self, url: Union[str, "HttpUrl"], filename: str = "", folder: Path = "") -> Path:
		self.__session.headers["User-Agent"] = self.private.headers["User-Agent"]
		self.__session.headers["Accept-Language"] = self.private.headers["Accept-Language"]
		#url = self.sanitize_instagram_url(url)
		fname = urlparse(str(url)).path.rsplit("/", 1)[1]
		filename = f"{filename}.{fname.rsplit('.', 1)[1]}" if filename else fname
		path = Path(folder or Path.cwd()) / filename

		logging.info("Downloading video from '%s' to '%s'", url, path)

		prepared = self.__session.prepare_request(requests.Request("GET", url))
		logging.info("Prepared headers: %s", prepared.headers)
		response = self.__session.send(
			prepared,
			stream=True,
			verify=False,
			proxies=self.public.proxies,
			timeout=self.request_timeout
		)
		response.raise_for_status()
		logging.info("Response headers: %s", response.headers)

		content_length = 0
		try:
			content_length = int(response.headers.get("Content-Length", 0))
		except (TypeError, ValueError):
			logging.warning("Content-Length header is missing or invalid.")

		downloaded = 0
		with open(path, "wb") as f:
			for chunk in response.iter_content(chunk_size=self.adaptive_chunk_size(content_length)):
				if chunk:
					f.write(chunk)
					downloaded += len(chunk)
					if self.progress_callback:
						try:
							self.progress_callback(
								total=content_length or None,
								bytes_transferred=downloaded,
								path=path
							)
						except Exception as e:
							logging.warning("Progress callback raised an exception!", exc_info=e)


		if content_length and downloaded != content_length:
			raise VideoNotDownload(
				f'Broken file "{path}" (expected {content_length}, got {downloaded})'
			)

		return path.resolve()

	def photo_download_by_url(
		self, url: Union[str, "HttpUrl"], filename: str = "", folder: Path = ""
	) -> Path:
		self.__session.headers["User-Agent"] = self.private.headers["User-Agent"]
		self.__session.headers["Accept-Language"] = self.private.headers["Accept-Language"]
		#url = self.sanitize_instagram_url(url)
		fname = urlparse(str(url)).path.rsplit("/", 1)[1]
		#filename = f"{filename}.{(filename, fname.rsplit('.', 1)[1]) if filename else fname}"
		filename = f"{filename}.{fname.rsplit('.', 1)[1]}" if filename else fname
		path = Path(folder) / filename

		logging.info("Downloading photo from '%s' to '%s'", url, path)
		logging.info("[Downloader] Using proxies: %s", self.public.proxies)

		prepared = self.__session.prepare_request(requests.Request("GET", url))
		logging.info("Prepared headers: %s", prepared.headers)
		response = self.__session.send(
			prepared,
			stream=True,
			verify=False,
			proxies=self.public.proxies,
			timeout=self.request_timeout
		)
		response.raise_for_status()
		logging.info("Response headers: %s", response.headers)

		content_length = 0
		try:
			content_length = int(
				response.headers.get("x-full-image-content-length")
				or response.headers.get("Content-Length")
				or 0
			)
		except (TypeError, ValueError):
			logging.warning("Content-Length header is missing or invalid.")

		downloaded = 0
		with open(path, "wb") as f:
			response.raw.decode_content = True
			for chunk in response.iter_content(chunk_size=self.adaptive_chunk_size(content_length)):
				if chunk:
					f.write(chunk)
					downloaded += len(chunk)
					if self.progress_callback:
						try:
							self.progress_callback(
								total=content_length or None,
								bytes_transferred=downloaded,
								path=path
							)
						except Exception as e:
							logging.warning("Progress callback raised an exception!", exc_info=e)
		return path.resolve()
	
	def media_info(self, media_pk: str, use_cache: bool = True) -> Media:
		"""
		Get Media Information from PK

		Parameters
		----------
		media_pk: str
			Unique identifier of the media
		use_cache: bool, optional
			Whether or not to use information from cache, default value is True

		Returns
		-------
		Media
			An object of Media type
		"""
		media_pk = self.media_pk(media_pk)
		if not use_cache or media_pk not in self._medias_cache:
			media = self.media_info_v1(media_pk)
			self._medias_cache[media_pk] = media
		return deepcopy(
			self._medias_cache[media_pk]
		)  # return copy of cache (dict changes protection)

	def media_info_v1(self, media_pk: str) -> Media:
		"""
		Get Media from PK by Private Mobile API

		Parameters
		----------
		media_pk: str
			Unique identifier of the media

		Returns
		-------
		Media
			An object of Media type
		"""
		try:
			result = self.private_request(f"media/{media_pk}/info/")
		except ClientNotFoundError as e:
			raise MediaNotFound(e, media_pk=media_pk, **self.last_json)
		except ClientError as e:
			if "Media not found" in str(e):
				raise MediaNotFound(e, media_pk=media_pk, **self.last_json)
			raise e
		return safe_extract_media_v1(result["items"].pop())
	
	def user_info_by_username(self, username: str, use_cache: bool = True) -> User:
		"""
		Get user object from username

		Parameters
		----------
		username: str
			User name of an instagram account
		use_cache: bool, optional
			Whether or not to use information from cache, default value is True

		Returns
		-------
		User
			An object of User type
		"""
		username = str(username).lower()
		if not use_cache or username not in self._usernames_cache:
			user = self.user_info_by_username_v1(username)
			self._users_cache[user.pk] = user
			self._usernames_cache[user.username] = user.pk
		return self.user_info(self._usernames_cache[username])
	
	def user_info(self, user_id: str, use_cache: bool = True) -> User:
		"""
		Get user object from user id

		Parameters
		----------
		user_id: str
			User id of an instagram account
		use_cache: bool, optional
			Whether or not to use information from cache, default value is True

		Returns
		-------
		User
			An object of User type
		"""
		user_id = str(user_id)
		if not use_cache or user_id not in self._users_cache:
			user = self.user_info_v1(user_id)
			self._users_cache[user_id] = user
			self._usernames_cache[user.username] = user.pk
		return deepcopy(
			self._users_cache[user_id]
		)  # return copy of cache (dict changes protection)
	
	def user_medias(self, user_id: str, amount: int = 0, sleep: int = 0) -> List[Media]:
		"""
		Get a user's media

		Parameters
		----------
		user_id: str
		amount: int, optional
			Maximum number of media to return, default is 0 (all medias)
		sleep: int, optional
			Timeout between page iterations

		Returns
		-------
		List[Media]
			A list of objects of Media
		"""
		amount = int(amount)
		user_id = int(user_id)
		sleep = int(sleep)
		# User may been private, attempt via Private API
		# (You can check is_private, but there may be other reasons,
		#  it is better to try through a Private API)
		medias = self.user_medias_v1(user_id, amount, sleep)
		return medias
	
	def user_medias_v1(self, user_id: str, amount: int = 0, sleep: int = 0) -> List[Media]:
		"""
		Get a user's media by Private Mobile API

		Parameters
		----------
		user_id: str
		amount: int, optional
			Maximum number of media to return, default is 0 (all medias)

		Returns
		-------
		List[Media]
			A list of objects of Media
		"""
		amount = int(amount)
		user_id = int(user_id)
		medias = []
		next_max_id = ""
		while True:
			try:
				medias_page, next_max_id = self.user_medias_paginated_v1(
					user_id, amount, end_cursor=next_max_id
				)
			except PrivateError as e:
				raise e
			except Exception as e:
				self.logger.exception(e)
				break
			medias.extend(medias_page)
			if not next_max_id:
				break
			if amount and len(medias) >= amount:
				break
			if sleep:
				time.sleep(sleep)
		if amount:
			medias = medias[:amount]
		return medias

	def user_stories(self, user_id: str, amount: int = None) -> List[Story]:
		"""
		Get a user's stories

		Parameters
		----------
		user_id: str
		amount: int, optional
			Maximum number of story to return, default is all

		Returns
		-------
		List[Story]
			A list of objects of STory
		"""
		return self.user_stories_v1(user_id, amount)


#from typing import TypedDict
import copy
from typing_extensions import Unpack

from warp_beacon.jobs.upload_job import UploadJob
from warp_beacon.jobs.abstract import AbstractJob, JobSettings

#import logging

class DownloadJob(AbstractJob):
	def __init__(self, **kwargs: Unpack[JobSettings]) -> None:
		super(DownloadJob, self).__init__(**kwargs)

	@staticmethod
	def build(**kwargs: Unpack[JobSettings]) -> "DownloadJob":
		return DownloadJob(**kwargs)
	
	def to_upload_job(self, **kwargs: Unpack[JobSettings]) -> AbstractJob:
		d = self.to_dict()
		d.update(copy.deepcopy(kwargs))
		if "media_collection" in d:
			for index, _ in enumerate(d["media_collection"]):
				for k, _ in enumerate(d["media_collection"][index]):
					d["media_collection"][index][k] = UploadJob.build(**d["media_collection"][index][k])
		return UploadJob.build(**d)


import time
import random
from typing import Optional
from datetime import datetime

import logging

from instagrapi.types import UserShort, Story
from warp_beacon.scraper.instagram.instagram import InstagramScraper

class InstagramHuman(object):
	default_profiles = ["nasa", "natgeo", "9gag", "spotify", "nba"]

	def __init__(self, scrapler: InstagramScraper) -> None:
		self.scrapler = scrapler
		self.operations_count = 0
		self.was_profile_explore = False
		self.tray_feed = None

	def reel_tray_feed_if_needed(self, reason: str = "pull_to_refresh") -> dict:
		if self.tray_feed is None:
			self.tray_feed = self.scrapler.download_hndlr(self.scrapler.cl.get_reels_tray_feed, reason)
			self.operations_count += 1

		return self.tray_feed

	def browse_timeline(self) -> Optional[dict]:
		feed = None
		items = []
		try:
			reason = random.choice(["cold_start_fetch", "pull_to_refresh"])
			feed = self.scrapler.cl.get_timeline_feed(reason=reason)
			self.operations_count += 1
			items = feed.get("feed_items", [])
		except Exception as e:
			logging.warning("Failed to get timeline feed!", exc_info=e)
			return

		seen = []
		if items:
			for item in items:
				media = item.get("media_or_ad")
				logging.debug("Item content: %s", media)
				if not media:
					continue
				media_id = media.get("id")
				#user_id = media.get("user", {}).get("pk")

				if media_id:
					seen.append(str(media_id))
					if random.random() > 0.8:
						try:
							#self.scrapler.cl.media_like(media_id)
							self.scrapler.cl.media_comments(media_id)
							self.operations_count += 1

							self.random_pause(short=True)
						except Exception as e:
							logging.warning("Failed to see comments to media '%s'", media_id, exc_info=e)
					self.random_pause()
		if seen:
			try:
				self.scrapler.cl.media_seen(seen)
				self.operations_count += 1
			except Exception as e:
				logging.warning("Failed to mark timeline feed as seen", exc_info=e)

		return feed

	def watch_stories(self) -> None:
		logging.info("Simulating stories watch ...")
		stories: list[Story] = []
		try:
			raw_tray = self.reel_tray_feed_if_needed()
			if not raw_tray:
				logging.info("No stories tray available.")
				return
			#logging.info("raw_tray: %s", str(raw_tray))
			tray = raw_tray.get("tray", [])
			filtered_tray = [item for item in tray if item.get("seen", 0) == 0]
			if filtered_tray:
				reels_tray = []
				target_len = random.randint(1, 5)
				for el in filtered_tray[:target_len]:
					user = el["user"]
					user_id = user["pk"]
					reels_tray.append({"user_id": user_id})

				for el in reels_tray:
					# amount?
					_stories = self.scrapler.download_hndlr(self.scrapler.cl.user_stories, el["user_id"])
					self.operations_count += 1
					if _stories:
						stories.extend(_stories)
					self.random_pause()

			if not stories:
				return

			seen = []
			explore_user = None
			if stories:
				for m in stories[:random.randint(1, len(stories))]:
					logging.info("Wathing story with pk '%s'", str(m.id))
					seen.append(str(m.id))
					if random.random() > 0.9:
						explore_user = m.user
						self.operations_count += 1
						break
					pause = 0.0
					if m.media_type == 2:
						pause = m.video_duration or random.uniform(2, 20)
					elif m.media_type == 8:
						pause = random.uniform(2, 31)
					else: # img, etc.
						pause = random.uniform(2, 14)
					logging.info("Pause for '%.2f' sec ...", round(pause, 2))
					time.sleep(pause)
					#self.random_pause()

			if seen:
				self.scrapler.download_hndlr(self.scrapler.cl.media_seen, seen)
				self.operations_count += 1
				logging.info("Marked '%d' stories as seen", len(seen))

			if explore_user:
				self.explore_profile(explore_user)
		except Exception as e:
			logging.warning("Failed to get user stories!", exc_info=e)

	def watch_content(self, media: list) -> None:
		if not media:
			return
		explore_user = None
		seen = []
		random.seed(time.time())
		for m in media[:random.randint(1, len(media))]:
			logging.info("Watching content with pk '%s'", str(m.id))
			seen.append(str(m.id))
			logging.info("Watched content with id '%s'", str(m.id))
			if random.random() > 0.9:
				explore_user = m.user
				break
			self.random_pause()

		self.scrapler.download_hndlr(self.scrapler.cl.media_seen, seen)
		self.operations_count += 1

		if explore_user:
			self.explore_profile(explore_user)

	def scroll_content(self, last_pk: int) -> None:
		#timeline_initialized = False
		if random.random() > 0.95:
			#timeline_initialized = True
			#self.browse_timeline()
			logging.info("Starting to watch related reels with media_pk '%d'", last_pk)
			media = self.scrapler.download_hndlr(self.scrapler.cl.reels, amount=random.randint(4, 10), last_media_pk=last_pk)
			self.operations_count += 1
			self.watch_content(media)
		
		if random.random() > 0.95:
			time.sleep(random.uniform(2, 20))
			#if not timeline_initialized:
			#	self.browse_timeline()
			logging.info("Starting to explore reels with media_pk '%d'", last_pk)
			media = self.scrapler.download_hndlr(self.scrapler.cl.explore_reels, amount=random.randint(4, 10), last_media_pk=last_pk)
			self.operations_count += 1
			self.watch_content(media)

	def simulate_activity(self) -> None:
		now = datetime.now()
		hour = now.hour

		if 6 <= hour < 11:
			self.morning_routine()
		elif 11 <= hour < 18:
			self.daytime_routine()
		elif 18 <= hour < 23:
			self.evening_routine()
		else:
			self.night_routine()

	def morning_routine(self) -> None:
		try:
			logging.info("Starting morning activity simulation")
			#self.scrapler.timeline_cursor = self.scrapler.download_hndlr(self.scrapler.cl.get_timeline_feed, "pull_to_refresh", self.scrapler.timeline_cursor.get("next_max_id"))
			#self.operations_count += 1
			self.browse_timeline()
			time.sleep(random.uniform(3, 7))
			if random.random() > 0.5:
				self.check_direct()
			if random.random() > 0.6:
				self.scrapler.download_hndlr(self.scrapler.cl.notification_like_and_comment_on_photo_user_tagged, "everyone")
				self.operations_count += 1
				self.random_pause()
			if random.random() > 0.2:
				logging.info("Simulation updating reels tray feed ...")
				self.reel_tray_feed_if_needed()
				self.random_pause()
			if random.random() > 0.4:
				self.watch_stories()
			if random.random() > 0.8:
				self.profile_view()
		except Exception as e:
			logging.warning("Error in morning_routine")
			logging.exception(e)

	def daytime_routine(self) -> None:
		try:
			logging.info("Starting day fast check activity simulation")
			#self.scrapler.download_hndlr(self.scrapler.cl.get_timeline_feed, "pull_to_refresh")
			#self.operations_count += 1
			self.browse_timeline()
			time.sleep(random.uniform(2, 5))
			if random.random() > 0.5:
				self.reel_tray_feed_if_needed()
				self.random_pause()

			if random.random() > 0.4:
				self.watch_stories()
				self.random_pause()

			if random.random() > 0.4:
				logging.info("Watching reels ...")
				reels = self.scrapler.download_hndlr(self.scrapler.cl.reels, amount=random.randint(4, 15))
				self.operations_count += 1
				self.watch_content(reels)
				self.random_pause()
		except Exception as e:
			logging.warning("Error in daytime_routine")
			logging.exception(e)

	def evening_routine(self) -> None:
		try:
			logging.info("Starting evening active user simulation")
			self.browse_timeline()
			time.sleep(random.uniform(2, 5))
			if random.random() > 0.5:
				self.check_direct()
			if random.random() > 0.6:
				logging.info("Checking notifications, tags ...")
				self.scrapler.download_hndlr(self.scrapler.cl.notification_like_and_comment_on_photo_user_tagged, "everyone")
				self.operations_count += 1
				self.random_pause()
			if random.random() > 0.4:
				self.watch_stories()
				self.random_pause()
			if random.random() > 0.4:
				logging.info("Watching reels ...")
				reels = self.scrapler.download_hndlr(self.scrapler.cl.reels, amount=random.randint(4, 10))
				self.operations_count += 1
				self.watch_content(reels)
				self.random_pause()
			if random.random() > 0.4:
				logging.info("Simulation profile view ...")
				self.profile_view()
				self.random_pause()
		except Exception as e:
			logging.warning("Error in evening_routine")
			logging.exception(e)

	def night_routine(self) -> None:
		try:
			logging.info("Starting night activity simulation")
			if random.random() > 0.7:
				self.check_direct()
			if random.random() > 0.7:
				self.watch_stories()
				self.random_pause()
			if random.random() > 0.5:
				logging.info("Watching reels ...")
				reels = self.scrapler.download_hndlr(self.scrapler.cl.reels, amount=random.randint(4, 15))
				self.operations_count += 1
				self.watch_content(reels)
				self.random_pause()
		except Exception as e:
			logging.warning("Error in night_routine")
			logging.exception(e)

	def random_pause(self, short: bool=False) -> None:
		pause = random.uniform(3, 10) if short else random.uniform(10, 30)
		logging.info("Pause for '%.2f' sec ...", round(pause, 2))
		time.sleep(pause)

	def check_direct(self) -> None:
		logging.info("Checking direct ...")
		self.scrapler.download_hndlr(self.scrapler.cl.direct_active_presence)
		self.operations_count += 1
		self.random_pause()
		threads = self.scrapler.download_hndlr(self.scrapler.cl.direct_threads, amount=random.randint(3, 7))
		self.operations_count += 1
		for thread in threads:
			messages = self.scrapler.cl.direct_messages(thread.id, amount=random.randint(5, 15))
			self.operations_count += 1
			if not messages:
				continue
			#msg_sample = random.sample(messages, k=random.randint(1, min(len(messages), 5)))
			#for msg in msg_sample:
				#if random.random() < 0.85:
					#self.scrapler.cl.direct_message_seen(msg.thread_id, msg.id)
					#self.operations_count += 1
					#logging.info("visual_media: '%s'", msg.visual_media)
				#self.random_pause()
			#self.random_pause()

	def explore_profile(self, user: UserShort) -> None:
		try:
			if self.was_profile_explore:
				return
			logging.info("Exploring user profile '%s'", user.username)
			self.was_profile_explore = True
			user_id = self.scrapler.download_hndlr(self.scrapler.cl.user_id_from_username, user.username)
			self.operations_count += 1
			if user_id:
				user_medias = self.scrapler.download_hndlr(self.scrapler.cl.user_medias, user_id, amount=random.randint(5, 15), sleep=random.randint(3, 7))
				self.operations_count += 1
				if user_medias:
					self.random_pause(short=True)
					self.watch_content(user_medias)
			if random.random() > 0.7:
				self.random_pause()
				logging.info("Watching '%s' followers ...", user.username)
				followers_amount = random.choices([3, 5, 8, 12], weights=[0.4, 0.3, 0.2, 0.1])[0]
				self.scrapler.download_hndlr(self.scrapler.cl.user_followers, user.pk, amount=followers_amount)
				self.operations_count += 1
				self.random_pause(short=True)
		except Exception as e:
			logging.warning("Failed to explore user profile with username '%s'", user.username, exc_info=e)

	def profile_view(self) -> None:
		try:
			logging.info("profile_view ...")
			my_user_id = self.scrapler.cl.user_id
			logging.info("user_following ...")
			friends = list(self.scrapler.download_hndlr(self.scrapler.cl.user_following, my_user_id, amount=random.randint(5, 15)).values())
			self.operations_count += 1
			time.sleep(random.uniform(2, 5))
			if not friends:
				friends = self.default_profiles
			
			random_friend = random.choice(friends)
			target_user_id = ""
			if isinstance(random_friend, UserShort):
				target_user_id = random_friend.pk
				logging.info("user_info with target_user_id = '%s' ...", target_user_id)
				self.scrapler.download_hndlr(self.scrapler.cl.user_info, target_user_id)
				#self.scrapler.download_hndlr(self.scrapler.cl.user_info_v1, target_user_id)
				self.operations_count += 1
				self.random_pause()
			elif isinstance(random_friend, str):
				target_user_id = self.scrapler.download_hndlr(self.scrapler.cl.user_id_from_username, random_friend)
				logging.info("user_info with target_user_id = '%s' ...", target_user_id)
				self.scrapler.download_hndlr(self.scrapler.cl.user_info, target_user_id)
				#self.scrapler.download_hndlr(self.scrapler.cl.user_info_v1, target_user_id)
				self.operations_count += 1
				self.random_pause()
			
			#self.scrapler.cl.explore_page_media_info

			if random.random() > 0.5:
				self.check_direct()

			if random.random() > 0.3:
				logging.info("Checking notifications, tags ...")
				self.scrapler.download_hndlr(self.scrapler.cl.notification_like_and_comment_on_photo_user_tagged, "everyone")
				self.operations_count += 1
				self.random_pause()
			
			if random.random() > 0.5:
				logging.info("user_medias with target_user_id = '%s' ...", target_user_id)
				self.scrapler.download_hndlr(self.scrapler.cl.user_medias_v1, target_user_id, amount=random.randint(1, 10))
				self.operations_count += 1
				self.random_pause()
		except Exception as e:
			logging.warning("Error in profile view")
			logging.exception(e)



import asyncio
import re

import logging

from pyrogram.client import Client
from pyrogram.types import InputMedia, InputMediaAudio, InputMediaPhoto, InputMediaVideo, InputMediaAnimation, InlineKeyboardMarkup
from pyrogram import raw
from pyrogram import types
from pyrogram.errors import FloodWait

from warp_beacon.telegram.progress_bar import ProgressBar
from warp_beacon.telegram.types import ReportType

class EditMessage(object):
	def __init__(self, client: Client) -> None:
		self.client = client
		self.base64_str_tpl = re.compile(r"[A-Za-z0-9\-_]{30,}")

	def get_wrapped_video(self, raw_file: raw.base.InputFile, raw_thumb: raw.base.InputFile, media: InputMediaVideo, file_name: str = None) -> raw.types.InputMediaUploadedDocument:
		return raw.types.InputMediaUploadedDocument(
				file=raw_file,
				mime_type=self.client.guess_mime_type(media.media) or "video/mp4",
				thumb=raw_thumb,
				spoiler=media.has_spoiler,
				attributes=[
					raw.types.DocumentAttributeVideo(
						duration=media.duration,
						w=media.width,
						h=media.height,
						supports_streaming=media.supports_streaming
					),
					raw.types.DocumentAttributeFilename(
						file_name=file_name
					)
				]
			)
	
	def get_wrapped_photo(self, raw_file: raw.base.InputFile, media: InputMediaPhoto) -> raw.types.InputMediaUploadedPhoto:
		return raw.types.InputMediaUploadedPhoto(
			file=raw_file,
			spoiler=media.has_spoiler
		)
	
	def get_wrapped_audio(self, raw_file: raw.base.InputFile, raw_thumb: raw.base.InputFile, media: InputMediaAudio, file_name: str = None) -> raw.types.InputMediaUploadedDocument:
		return raw.types.InputMediaUploadedDocument(
			mime_type=self.client.guess_mime_type(media.media) or "audio/mpeg",
			thumb=raw_thumb,
			file=raw_file,
			attributes=[
				raw.types.DocumentAttributeAudio(
					duration=media.duration,
					performer=media.performer,
					title=media.title
				),
				raw.types.DocumentAttributeFilename(
					file_name=file_name
				)
			]
		)
	
	def get_wrapped_animation(self, raw_file: raw.base.InputFile, raw_thumb: raw.base.InputFile, media: InputMediaVideo, file_name: str = None) -> raw.types.InputMediaUploadedDocument:
		return raw.types.InputMediaUploadedDocument(
			mime_type=self.client.guess_mime_type(media.media) or "video/mp4",
			thumb=raw_thumb,
			spoiler=media.has_spoiler,
			file=raw_file,
			attributes=[
				raw.types.DocumentAttributeVideo(
					supports_streaming=True,
					duration=media.duration,
					w=media.width,
					h=media.height
				),
				raw.types.DocumentAttributeFilename(
					file_name=file_name
				),
				raw.types.DocumentAttributeAnimated()
			]
		)

	def looks_like_token(self, s: str) -> bool:
		return bool(re.fullmatch(self.base64_str_tpl, s))

	async def edit(self,
		chat_id: int | str,
		message_id: int,
		media: InputMedia | InputMediaAudio | InputMediaPhoto | InputMediaVideo | InputMediaAnimation,
		reply_markup: InlineKeyboardMarkup = None,
		file_name: str = None
	) -> None:
		is_looks_like_token = self.looks_like_token(media.media)
		raw_file = None
		if not is_looks_like_token:
			progress_bar = ProgressBar(self.client)
			#await progress_bar.progress_callback(current=0, total=0, chat_id=chat_id, message_id=message_id, operation="Uploading", label=file_name, report_type=ReportType.PROGRESS)
			raw_file = await self.client.save_file(path=media.media, progress=progress_bar.progress_callback, progress_args=(chat_id, message_id, "Uploading", ReportType.PROGRESS, file_name))
		else:
			raw_file = media.media

		caption = media.caption
		parse_mode = media.parse_mode

		message, entities = None, None

		if caption is not None:
			message, entities = (await self.client.parser.parse(caption, parse_mode)).values()

		raw_media = None
		if isinstance(media, types.InputMediaVideo):
			#progress_bar_thumb = ProgressBar(self.client)
			raw_file_thumb = None
			if not is_looks_like_token:
				raw_file_thumb = await self.client.save_file(path=media.thumb)
			raw_media = self.get_wrapped_video(raw_file=raw_file, raw_thumb=raw_file_thumb, media=media, file_name=file_name)
		elif isinstance(media, types.InputMediaPhoto):
			raw_media = self.get_wrapped_photo(raw_file=raw_file, media=media)
		elif isinstance(media, types.InputMediaAudio):
			#progress_bar_thumb = ProgressBar(self.client)
			raw_file_thumb = None
			if not is_looks_like_token:
				raw_file_thumb = await self.client.save_file(path=media.thumb)
			raw_media = self.get_wrapped_audio(raw_file=raw_file, raw_thumb=raw_file_thumb, media=media, file_name=file_name)
		elif isinstance(media, types.InputMediaAnimation):
			#progress_bar_thumb = ProgressBar(self.client)
			raw_file_thumb = None
			if not is_looks_like_token:
				raw_file_thumb = await self.client.save_file(path=media.thumb)
			raw_media = self.get_wrapped_animation(raw_file=raw_file, raw_thumb=raw_file_thumb, media=media, file_name=file_name)

		peer, r = None, None
		while True:
			try:
				peer = await self.client.resolve_peer(chat_id)
				r = await self.client.invoke(
					raw.functions.messages.EditMessage(
						peer=peer,
						id=message_id,
						media=raw_media,
						reply_markup=await reply_markup.write(self.client) if reply_markup else None,
						message=message,
						entities=entities
					)
				)
				break
			except FloodWait as e:
				logging.warning("FloodWait occurred, waiting '%d' seconds before retry", int(e.value))
				await asyncio.sleep(float(e.value))

		if r:
			for i in r.updates:
				if isinstance(i, (raw.types.UpdateEditMessage, raw.types.UpdateEditChannelMessage)):
					return await types.Message._parse(
						self.client, i.message,
						{i.id: i for i in r.users},
						{i.id: i for i in r.chats}
					)


import os
import sys
from pathlib import Path
import importlib.util
import logging

from pyrogram import Client
from pyrogram.types import Message

class CustomHandlers(object):
	def __init__(self) -> None:
		self.handlers = []
		self.ch_dir = os.environ.get("CUSTOM_HANDLERS_DIR", default="/var/warp_beacon/custom_handlers")
		try:
			if self.ch_dir and os.path.isdir(self.ch_dir):
				for f in os.listdir(self.ch_dir):
					handler_path = f"{self.ch_dir}/{f}"
					if f.endswith(".py"):
						logging.info("[CUSTOM HANDLERS]: Found file '%s'", handler_path)
						handler = self.load_class_from_file(file_path=handler_path)
						self.handlers.append(handler)
			else:
				logging.info("[CUSTOM HANDLERS]: Directory '%s' is not exists, skipping custom handlers init", self.ch_dir)
		except Exception as e:
			logging.warning("[CUSTOM HANDLERS]: Exception occurred while custom handlers init!", exc_info=e)

	def load_class_from_file(self, file_path: str):
		module_name = Path(file_path).stem # module name from file name
		class_name = module_name
		spec = importlib.util.spec_from_file_location(module_name, file_path)
		module = importlib.util.module_from_spec(spec)
		sys.modules[module_name] = module
		spec.loader.exec_module(module)

		logging.info("[CUSTOM HANDLERS]: Loaded custom handler '%s'", class_name)

		cls = getattr(module, module_name, None)
		if isinstance(cls, type):
			return cls()

		raise RuntimeError(f"No handler found in {file_path}")
	
	async def exec_handlers(self, client: Client, message: Message, message_text: str) -> None:
		try:
			if not self.handlers:
				return
			for h in self.handlers:
				await h.exec(client, message, message_text)
		except Exception as e:
			logging.warning("[CUSTOM HANDLERS]: Exception occurred while running custom handler!", exc_info=e)


import pickle

import logging

from warp_beacon.storage.mongo import DBClient
from warp_beacon.jobs.download_job import DownloadJob

class FailHandler(object):
	client = None
	db = None
	def __init__(self, client: DBClient) -> None:
		self.client = client
		self.db = self.client.client.media.failed_jobs

	def __del__(self) -> None:
		self.client.close()

	def store_failed_job(self, job: DownloadJob) -> int:
		db_id = ""
		try:
			# check if job not already in storage
			find_opts = {"uniq_id": job.uniq_id, "message_id": job.message_id, "chat_id": job.chat_id}
			if self.db.find_one(find_opts) is not None:
				logging.info("Job already in storage skipping write")
				return 0
			#
			job_serilized = pickle.dumps(job)
			db_id = self.db.insert_one(
			{
				"job_data": job_serilized,
				"uniq_id": job.uniq_id,
				"message_id": job.message_id,
				"chat_id": job.chat_id
			}).inserted_id
		except Exception as e:
			logging.error("Failed to store job as failed!")
			logging.exception(e)
		return db_id
	
	def get_failed_jobs(self, clean: bool = True) -> list:
		ret = []
		try:
			cursor = self.db.find()
			for document in cursor:
				ret.append({
					"_id": document["_id"],
					"job": pickle.loads(document["job_data"]),
					"uniq_id": document.get("uniq_id"),
					"message_id": document.get("message_id"),
					"chat_id": document.get("chat_id")
				})
			if clean:
				self.db.delete_many({})
		except Exception as e:
			logging.error("Failed to get failed jobs!")
			logging.exception(e)
		return ret
	
	def remove_failed_job(self, uniq_id: str) -> bool:
		try:
			result = self.db.delete_one({"uniq_id": uniq_id})
			if result.deleted_count > 0:
				return True
		except Exception as e:
			logging.error("Failed to remove failed job!", exc_info=e)

		return False


import logging

from fake_useragent import UserAgent

class ScraperUtils(object):
	@staticmethod
	def get_ua_dict(browsers: list=['Facebook', 'Android'], platforms: list=['mobile', 'tablet'], os: list=['Android', 'iOS']) -> dict:
		random_client = None
		try:
			ua = UserAgent(browsers=browsers, platforms=platforms, os=os)
			random_client = ua.getRandom
			logging.info("Select random UA: %s", random_client)
		except Exception as e:
			logging.warning("Exception occurrd while generating random client UA!", exc_info=e)
			random_client = {'useragent': 'Mozilla/5.0 (Linux; Android 14; SM-S911B Build/UP1A.231005.007; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/133.0.6943.117 Mobile Safari/537.36 [FB_IAB/FB4A;FBAV/502.0.0.66.79;IABMV/1;]', 'percent': 0.017937771404345798, 'type': 'mobile', 'device_brand': 'Samsung', 'browser': 'Facebook', 'browser_version': '502.0.0', 'browser_version_major_minor': 502.0, 'os': 'Android', 'os_version': '14', 'platform': 'Linux aarch64'}
		return random_client

	@staticmethod
	def get_ua(browsers: list=['Facebook', 'Android'], platforms: list=['mobile', 'tablet'], os: list=['Android', 'iOS']) -> str:
		return ScraperUtils.get_ua_dict(browsers=browsers, platforms=platforms, os=os)["useragent"]


import os
import time
import logging
from mimetypes import guess_extension, guess_type
from urllib.parse import urlparse
import requests
import yt_dlp
from playwright.sync_api import sync_playwright, Page

from warp_beacon.telegram.utils import Utils
from warp_beacon.scraper.utils import ScraperUtils
from warp_beacon.jobs.types import JobType
from warp_beacon.scraper.X.abstract import XAbstract

from warp_beacon.scraper.exceptions import Unavailable

class XScraper(XAbstract):
	DOWNLOAD_DIR = "/tmp"

	def extract_canonical_name(self, media: dict) -> str:
		ret = ""
		try:
			if media.get("description"):
				ret = media["description"]
			elif media.get("title"):
				ret = media["title"]
		except Exception as e:
			logging.warning("Failed to extract canonical media name!")
			logging.exception(e)

		return ret

	def generate_result(self, local_files: list, job_type: JobType, canonical_name: str = "", performer: str = "") -> list:
		res = []
		if local_files:
			if job_type == JobType.COLLECTION:
				chunks = []
				for media_chunk in Utils.chunker(local_files, 10):
					chunk = []
					for media in media_chunk:
						mime_type, _ = guess_type(media)
						chunk.append({
							"local_media_path": self.rename_local_file(media),
							"canonical_name": canonical_name,
							"media_type": JobType.VIDEO if "video" in mime_type else JobType.IMAGE,
							"media_info": {}
						})
					chunks.append(chunk)
					res.append({
						"media_type": JobType.COLLECTION,
						"canonical_name": canonical_name,
						"items": chunks
					})
			else:
				for local_file in local_files:
					res.append({
						"local_media_path": self.rename_local_file(local_file),
						"performer": performer,
						"canonical_name": canonical_name,
						"media_type": job_type
					})
		logging.debug(res)
		return res

	def _download(self, url: str, timeout: int = 60) -> list:
		res = []
		post_text = ""
		pw_proxy = None
		if self.proxy:
			dsn = self.proxy.get("dsn", "")
			if dsn:
				parsed = urlparse(dsn)
				pw_proxy = {
					"server": f"{parsed.scheme}://{parsed.hostname}:{parsed.port}",
					"username": parsed.username,
					"password": parsed.password
				}
				logging.info("[X] build proxy: %s", pw_proxy)

		contains_images, contains_videos = False, False
		images, videos = [], []
		with sync_playwright() as p:
			with p.chromium.launch(headless=True) as browser:
				with browser.new_context(proxy=pw_proxy, ignore_https_errors=True) as context:
					page = context.new_page()
					page.goto(url, wait_until="networkidle", timeout=(timeout*1000))
					page.wait_for_selector("article[role='article']", timeout=(timeout*1000))

					contains_videos = self.tweet_contains_video(page)
					contains_images = self.tweet_contains_images(page)

					if contains_images:
						post_text, images = self.download_images(page, timeout)

					if not contains_images and not contains_videos:
						post_text = self.extract_post_text(page)

		if contains_videos:
			media_info, videos = self.download_videos(url, timeout)
			if media_info:
				post_text = self.extract_canonical_name(media_info)

		if not images and not videos:
			if not post_text:
				raise Unavailable("Content unvailable")
			logging.info("[X]: Sending text message")
			res.append({
				"message_text": post_text,
				"media_type": JobType.TEXT
			})
			return res
		
		if len(images) > 1 or len(videos) > 1:
			logging.info("[X]: uploading collection")
			content = images + videos
			res.extend(self.generate_result(content, JobType.COLLECTION, canonical_name=post_text))
		else:
			logging.info("[X]: uploading media")
			for job_type, content in {JobType.IMAGE: images, JobType.VIDEO: videos}.items():
				if content:
					res.extend(self.generate_result(content, job_type, canonical_name=post_text))

		return res

	def download_videos(self, url: str, timeout: int = 60) -> tuple[dict, list[str]]:
		local_files = []
		media_info = {}
		time_name = str(time.time()).replace('.', '_')
		ydl_opts = {
			'socket_timeout': timeout,
			'outtmpl': f'{self.DOWNLOAD_DIR}/x_download_{time_name}_%(id)s.%(ext)s',
			'quiet': False,
			'force_generic_extractor': False,
			#'noplaylist': True,
			'merge_output_format': 'mp4',
			'dump_single_json': False,
			'nocheckcertificate': True,
			'progress_hooks': [self.dlp_on_progress],
		}
		if self.proxy:
			proxy_dsn = self.proxy.get("dsn", "")
			logging.info("[X] Using proxy DSN '%s'", proxy_dsn)
			if proxy_dsn:
				ydl_opts["proxy"] = proxy_dsn

		with yt_dlp.YoutubeDL(ydl_opts) as ydl:
			info = ydl.extract_info(url, download=False)
			media_info = info
			entries = info.get("entries", [info])

			for entry in entries:
				ret = ydl.download([entry['webpage_url']])
				if ret == 0:
					file_path = ydl.prepare_filename(entry)
					if isinstance(file_path, str):
						local_files.append(file_path)
					else:
						local_files.extend(file_path)

		return media_info, local_files

	def adaptive_chunk_size(self, content_length: int) -> int:
		if content_length < 100_000:
			return 2048
		elif content_length < 5_000_000:
			return 8192
		elif content_length < 100_000_000:
			return 32768
		else:
			return 65536

	def get_extension_from_headers(self, response: requests.Response) -> str:
		content_type = response.headers.get("Content-Type", "")
		return guess_extension(content_type) or ".jpg"

	def download_images(self, page: Page, timeout: int) -> tuple[str, list[str]]:
		downloaded_imgs = []
		headers = {
			"User-Agent": ScraperUtils.get_ua(),
			"Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
			"Accept-Language": "en-us,en;q=0.5",
			"Sec-Fetch-Mode": "navigate"
		}
		proxies = None
		if self.proxy:
			proxies = {"https": self.proxy.get("dsn", ""), "http": self.proxy.get("dsn", "")}

		image_urls, post_text = self.extract_image_urls_from_x_post(page, timeout)

		if not image_urls:
			logging.error("[X] Content images are not found!")
			return downloaded_imgs

		time_name = str(time.time()).replace('.', '_')
		for i, img_url in enumerate(set(image_urls)):
			downloaded = 0
			if "?name=small" in img_url:
				img_url = img_url.replace("?name=small", "?name=orig")
			with requests.get(
				img_url,
				headers=headers,
				timeout=timeout,
				stream=True,
				verify=False,
				proxies=proxies) as request:
				
				request.raise_for_status()

				parsed = urlparse(img_url)
				ext = os.path.splitext(parsed.path)[1]
				if not ext:
					ext = self.get_extension_from_headers(request)
				filename = f"x_download_{time_name}_{i}{ext}"
				filepath = os.path.join(self.DOWNLOAD_DIR, filename)

				content_length = int(request.headers.get("Content-Length", 0))

				with open(filepath, "wb") as f:
					#request.raw.decode_content = True
					chunk_size = self.adaptive_chunk_size(content_length)
					for chunk in request.iter_content(chunk_size=chunk_size):
						if chunk:
							f.write(chunk)
							downloaded += len(chunk)
							self.download_progress(
								total=content_length or None,
								bytes_transferred=downloaded,
								path=filepath
							)
				downloaded_imgs.append(filepath)

		return post_text, downloaded_imgs

	def extract_image_urls_from_x_post(self, page: Page, timeout: int) -> tuple[list[str], str]:
		img_urls, post_text = [], ''

		page.wait_for_selector("img[src*='pbs.twimg.com/media']", timeout=(timeout*1000))
		post_text = self.extract_post_text(page)

		image_elements = page.query_selector_all("img")
		image_urls = []

		for img in image_elements:
			src = img.get_attribute("src")
			if src and "pbs.twimg.com/media" in src:
				image_urls.append(src)
					
		img_urls = list(set(image_urls))
		return img_urls, post_text
	
	def tweet_contains_video(self, page: Page) -> bool:
		try:
			return bool(
				page.query_selector("article video") or
				page.query_selector("div[data-testid='videoPlayer']") or
				page.query_selector("div[aria-label='Embedded video']")
			)
		except Exception:
			pass
		return False
		
	def tweet_contains_images(self, page: Page) -> bool:
		try:
			image_elements = page.query_selector_all("img")
			image_urls = [
				img.get_attribute("src")
				for img in image_elements
				if img.get_attribute("src") and "pbs.twimg.com/media" in img.get_attribute("src")
			]
			return bool(image_urls)
		except Exception:
			pass
		return False
	
	def extract_post_text(self, page: Page) -> str:
		try:
			text_fragments = []

			# find tweetText containers (in main and quoted)
			containers = page.query_selector_all('div[data-testid="tweetText"]')
			for container in containers:
				fragments = []

				# find <span> and <img alt=...> inside text
				for node in container.query_selector_all("span, img"):
					tag = node.evaluate("node => node.tagName.toLowerCase()")
					if tag == "span":
						value = node.inner_text().strip()
						if value:
							fragments.append(value)
					elif tag == "img":
						# emoji as image
						alt = node.get_attribute("alt")
						if alt:
							fragments.append(alt)

				if fragments:
					text_fragments.append("".join(fragments))

			return "\n\n".join(text_fragments).strip()

		except Exception as e:
			logging.warning("X: [extract_post_text] error", exc_info=e)
		return ""